#!/usr/bin/perl

#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

my $len = 32;
my @chars = ( 'a' .. 'z', 'A' .. 'Z', '0' .. '9', '_' );
foreach my $i ( 0 .. 15 ) {
    my $string = "";
    foreach ( 1 .. $len ) {
        $string .= $chars[ rand @chars ];
    }
    print "key" . $i . " = " . $string . "\n";
}
#print "error_url=302 http://www.domain.com/this/is/the/path/error.html\n";
print "error_url = 403\n";
